
structure EDParse :> EDPARSE =
struct
    
    structure S = Stream
    structure M = EDMinML
    structure L = EDLexer

    exception ParseError of string

    fun I x = x

    fun next s =
        case S.force s of
            S.Nil => raise ParseError "Unexpected end of stream."
          | S.Cons result => result

    fun match tok s =
        let val (n, s) = next s
        in if tok = n then s else raise ParseError ("Expected " ^ EDLexer.tokenToString tok ^ 
                                                    " token")
        end

    fun primop L.PLUS = SOME M.Plus
      | primop L.EQUALS = SOME M.Equal
      | primop L.MINUS = SOME M.Minus
      | primop L.TIMES = SOME M.Times
      | primop L.NEGATE = SOME M.Negate
      | primop _ = NONE

    fun Prog r =
        let
            val (e, s) = Exp (S.delay (fn () => S.Cons r))
        in
            (e, match L.SEMICOLON s)
        end

    (* Exp recursively consumes adjacent atomic expressions (Expa),
       forming them into a chain of applications. *)

    and Exp es =
        let
            fun Exps s eo =
                case Exp_maybe s of
                    SOME (e, s) => (case eo of
                                        NONE => Exps s (SOME e)
                                      | SOME f => Exps s (SOME (M.Apply(f, e))))
                  | NONE => (case eo of
                                 NONE => raise ParseError "Expected Expression"
                               | SOME e => (e, s))
        in
            Exps es NONE
        end

    (* Expa : L.token * L.token S.stream -> (M.exp * L.token S.stream) option
       attempts to find an atomic expression (no applications) starting
       with the token, perhaps continuing through the stream. Returns
       SOME(e, s) if the exp e was successfully recognized, with s
       the stream remaining after it.
       Returns NONE if the token cannot begin any exp.
       May raise exception ParseError if the input stream does not represent
       any valid MinML program. *)
    and Expa (L.TRUE, s) = SOME (M.Bool true, s)
      | Expa (L.FALSE, s) = SOME (M.Bool false, s)
      | Expa (L.NUMBER n, s) = SOME (M.Int n, s)
      | Expa (L.VAR v, s) = SOME (M.Var v, s)
      | Expa (L.IF, s) =
          let
              val (ec, s) = Exp s
              val s = match L.THEN s
              val (et, s) = Exp s
              val s = match L.ELSE s
              val (ef, s) = Exp s
              val s = match L.FI s
          in
              SOME (M.If(ec, et, ef), s)
          end
      | Expa (L.LPAREN, s) =
          let
              val (e, s) = Exp s
          in
              SOME (e, match L.RPAREN s)
          end
      | Expa (L.FUN, s) =
          let
              val (f, s) = Var (next s)
              val s = match L.LPAREN s
              val (x, s) = Var (next s)
	      val (tok,s) = next s
	  in
	      case tok of
		  L.COLON =>
		  let
		      val (dt, s) = Type (next s)
		      val s = match L.RPAREN s
		      val s = match L.COLON s
		      val (rt, s) = Type (next s)
		      val s = match L.IS s
		      val (e, s) = Exp s
		  in
		      SOME (M.Fun(f, x, dt, rt, e), match L.END s)
		  end
		| L.RPAREN =>
		  let
		      val s = match L.IS s
		      val (e, s) = Exp s
		  in
		      SOME (M.UTFun(f,x,e), match L.END s)
		  end
	  end
      | Expa (L.CHECK,s) =
	let
	    val s = match L.LPAREN s
	    val (e,s) = Exp s
	    val s = match L.COMMA s
	    val (t,s) = Type (next s)
	    val s = match L.RPAREN s
	in SOME (M.Check(e,t),s)
	end
      | Expa (p, s) =
          (case primop p of
               SOME po =>
               let
                   val s = match L.LPAREN s
                   val (el, s) = ExpSeq s
               in
                   SOME (M.Primop (po, el), match L.RPAREN s)
               end
		   (* if it's not a primop, then no exp is possible *)
             | NONE => NONE)
	   
         
    (* XXX necessary? *)
    and Exp_maybe s = case S.force s of
        S.Nil => NONE
      | S.Cons res => Expa res
            
    and ExpSeq s =
        let
            fun es s =
                case (Exp s) of
                    (e, s) =>
                        (case next s of
                             (L.COMMA, s) =>
                                 let 
                                     val (el, s) = es s
                                 in
                                     (e :: el, s)
                                 end
                           | _ => ([e], s))
        in
            es s
        end

    and Var (L.VAR v, s) = (v, s)
      | Var _ = raise ParseError "expected Var"

    and Type (L.INT, s) = (M.INT, s)
      | Type (L.BOOL, s) = (M.BOOL, s)
      | Type (L.ARROW, s) =
        let
            val (d, s) = Type (next s)
            val (r, s) = Type (next s)
        in
            (M.ARROW(d, r), s)
        end
      | Type _ = raise ParseError "expected Type"

    (* exported parsing function *)
          
    fun parse (s : L.token S.stream) : M.exp S.stream = 
        S.delay (fn () => parse' (S.force s))

    and parse' S.Nil = S.Nil
      | parse' (S.Cons r) = 
        let 
            val (e, s) = Prog r
        in 
            S.Cons (e, parse s)
        end

end
